try:
    from math import prod
except ImportError:
    # very annoyed rn
    from functools import reduce
    from operator import mul
    def prod(it):
        return reduce(mul, it, 1)


from collections import namedtuple


SampleValue = namedtuple('SampleValue', ('effort', 'scaling_factor', 'gear', 'op_factor', 'dissipation'))


class SampleDetail(object):
    def __init__(self, effort, scaling_factor, gear, op_factor, dissipation):
        self.e_param, self.effort = effort
        self.g_param, self.gear = gear
        self.d_param, self.dissipation = dissipation
        self.scaling_factor = scaling_factor
        self.op_factor = op_factor
        self.gear_kg = self.effort * self.scaling_factor * self.gear * self.op_factor
        self.diss_kg = self.gear_kg * self.dissipation

    def __iter__(self):
        return iter(SampleValue(self.effort, self.scaling_factor, self.gear, self.op_factor, self.dissipation))


# SampleDetail = namedtuple('SampleDetail', ('effort', 'scaling_factor', 'gear', 'op_factor', 'dissipation'))


class ConflictingUnits(Exception):
    pass


class ConflictingParams(Exception):
    pass


class NoValidParams(Exception):
    pass


class ZeroValuedModel(Exception):
    pass


class GearModel(object):
    """
    Note on the directionality of unit conversions:

    Each stage is equipped with an "equivalence table" of quantities and their conversion factors with a unit of the
    fishing gear material flow. All the entries in an equivalence table correspond to the same amount, so units whose
    value is 1.0 are reference units.

    When two adjacent stages have different units, a conversion factor is applied. Traveling downstream, we divide by
    downstream equivalency factors or multiply by upstream ones.

    effort intensity: 0.4
    effort: vessel-set per catch    gear: mass per vessel meter    downstream: dissipation per year
    {set: 1.0                       {LOA: 1.0                      { year: 1.0
     hour: 6.0                       vessel:  0.04545                set: 120
     year: 0.005                    }                              }
    }
    A 22-meter vessel that catches 2.5 t per set has an effort intensity of 0.4 vessel-sets per tonne catch
    However, that same vessel has an effort intensity of 22*6 / 2.5 = 52.8 meter-hours per tonne.  Thus we multiply
    by the effort model's 6 hours per set equivalency and divide by the gear model's 0.0454545 vessels per meter.

    When we move on to dissipation per year, we get different results depending on which model we ask. The effort is
    consulted first (*0.005), but if the effort model doesn't know, the dissipation model is asked (/120)


    In the LCI case, the effort intensity is 1 ton of catch per ton of catch.  The gear intensity is direct: 580 g per
    ton, and some fraction of that 580 g is dissipated.  How much? the dissipation rate, per year. say 3%.  3% per year
    is, of course, 9% over the course of a gear's 3-year life, but for the year in question it's simply 3% of the mass.
    Thus

    n.b. I know family is not enough to uniquely identify models.

    """
    def __init__(self, e=None, g=None, d=None, label=None):
        self._e = e
        self._g = g
        self._d = d

        self.validate()

        self._label = label

    def validate(self):
        try:
            assert self.scaling_factor * self.op_factor != 0
        except AssertionError:
            raise ZeroValuedModel(self._e.family, self._g.family, self._d.family)

    @property
    def effort(self):
        return self._e

    @property
    def gear(self):
        return self._g

    @property
    def label(self):
        if self._label is None:
            fam = ['']
            for k in self.family:
                if k == fam[-1]:
                    continue
                fam.append(k)
            return '-'.join(fam[1:])
        return self._label

    @label.setter
    def label(self, title):
        self._label = str(title)

    @property
    def dissipation(self):
        return self._d

    def e_bar(self, param=None):
        return self._e.mean(param)

    def e_tilde(self, param=None):
        return self._e.sample(param)

    @property
    def scaling_unit(self):
        return self._e.scaling_unit

    def g_bar(self, param=None):
        return self._g.mean(param)

    def g_tilde(self, param=None):
        return self._g.sample(param)

    @property
    def op_unit(self):
        return self._e.op_unit

    def d_bar(self, param=None):
        return self._d.mean(param)

    def d_tilde(self, param=None):
        return self._d.sample(param)

    @property
    def family(self):
        return self._e.family, self._g.family, self._d.family

    @property
    def scaling_factor(self):
        """
        upstream is effort; downstream is gear. effort doesn't currently keep a scaling equiv (but should)
        :return:
        """

        if self._e.scaling_unit == self._g.scaling_unit:
            return 1.0
        elif self._e.scaling_unit in self._g.scaling_units:
            return 1.0 / self._g.scaling_factor(self._e.scaling_unit)
        else:
            raise ConflictingUnits('%s > %s' % (self._e.scaling_unit, list(self._g.scaling_units)))

    @property
    def op_factor(self):
        """
        upstream is effort; downstream is dissipation.
        :return:
        """
        if self._e.op_unit == self._d.op_unit:
            return 1.0
        elif self._d.op_unit in self._e.op_units:
            return self._e.op_factor(self._d.op_unit)
        elif self._e.op_unit in self._d.op_units:
            return 1.0 /  self._d.op_factor(self.op_unit)
        else:
            raise ConflictingUnits('%s > %s' % (list(self._e.op_units), list(self._d.op_units)))


    def _report_row(self, e_param, g_param, d_param):
        e, sf, g, of, d = self.mean_long(e_param=e_param, g_param=g_param, d_param=d_param)
        r = {
            'catch_family': self._e.family,
            'gear_family': self._g.family,
            'dissipation_family': self._d.family,
            'catch_unit': self._e.catch_unit.name,
            'scaling_unit': self.scaling_unit.name,
            'op_unit': self.op_unit.name,
            'effort_param': e_param,
            'effort': e,
            'scaling_factor': sf,
            'intensity_param': g_param,
            'intensity': g,
            'op_factor': of,
            'dissipation_op': self._d.op_unit.name,
            'dissipation_param': d_param,
            'dissipation': d,
            'dissipation_type': self._d.dissipation_type,
            'result': e*sf*g*of*d
        }

        r.update(self._g.attributes)
        return r

    def report(self, e_param=None, g_param=None, d_param=None):
        for ep in self._e.values(e_param):
            for gp in self._g.values(g_param):
                for dp in self._d.values(d_param):
                    yield self._report_row(ep, gp, dp)

    def mean_long(self, e_param=None, g_param=None, d_param=None):
        return SampleDetail(self.e_bar(e_param), self.scaling_factor, self.g_bar(g_param), self.op_factor, self.d_bar(d_param))

    def mean(self, e_param=None, g_param=None, d_param=None):
        return prod(self.mean_long(e_param=e_param, g_param=g_param, d_param=d_param))

    def sample_long(self, e_param=None, g_param=None, d_param=None):
        return SampleDetail(self.e_tilde(e_param), self.scaling_factor, self.g_tilde(g_param), self.op_factor, self.d_tilde(d_param))

    def sample(self, e_param=None, g_param=None, d_param=None):
        return prod(self.sample_long(e_param=e_param, g_param=g_param, d_param=d_param))

    def __str__(self):
        return "; ".join((self._e.name, self._g.name, self._d.name))

    @property
    def family(self):
        return self._e.family, self._g.family, self._d.family

    @property
    def formula(self):
        return "%s ~ %s * %s" % (self._e.catch_unit, self.scaling_unit, self.op_unit)

'''
class GearTraversal(GearModel):
    """
    What do you get when you traverse a gear model?
    """
    def __init__(self, gear_types, params=None, effort=None, gear=None, dissipation=None):
        super(GearTraversal, self).__init__()
        if params is None:
            params = dict()
        self.gear_ecs = 
        self.quals = set()
        self.quants = dict()

        for k, v in params.items():
            if v is True:
                self.quals.add(k)
            elif float(v) == v:
                self.quants[k] = v

        self.effort = effort
        self.gear = gear
        self.dissipation = dissipation

    def check_gear(self, mod):
        if self.gear_ecs.intersection(mod.gear_ecs):
            return True
        return False

    def _check_param(self, mod):
        if mod is None:
            return
        if mod.model_type == 'discrete':
            t = [k for k in self.quals if mod.is_valid(k)]
            if len(t) > 1:
                raise ConflictingParams()
            elif len(t) == 0:
                raise NoValidParams(list(mod.keys))
            return t[0]
        elif mod.model_type == 'continuous':
            if mod.order == 0:
                return None
            if mod.param_unit in self.quants:
                return self.quants[mod.param_unit]
            raise NoValidParams(mod.param_unit)

    @property
    def effort(self):
        return self._e

    @effort.setter
    def effort(self, mod):
        self._e_param = self._check_param(mod)
        self._e = mod

    @property
    def gear(self):
        return self._g

    @gear.setter
    def gear(self, mod):
        if mod is None:
            return
        if mod.scaling_unit != self.scaling_unit:
            raise ConflictingUnits('%s | %s' % (mod.scaling_unit, self.scaling_unit))
        self._g_param = self._check_param(mod)
        self._g = mod

    @property
    def dissipation(self):
        return self._d

    @dissipation.setter
    def dissipation(self, mod):
        if mod is None:
            return
        if mod.op_unit not in self._e.op_units and self._e.op_unit not in mod.op_units:
            raise ConflictingUnits('%s < %s' % (list(mod.op_units), list(self._e.op_units)))
        self._d_param = self._check_param(mod)
        self._d = mod

    def strip(self):
        return GearModel(self._e, self._e_param, self._g, self._g_param, self._d, self._d_param)

    def __str__(self):
        gm = next(list(self.gear_ecs)[0].values())
        return "%s: %s ~ %s%s" % (gm , self._e.catch_unit, self.scaling_unit, self.op_unit)
'''


"""
Note: from Gilman 2020-pre to Gilman 2021:
 - bottom trawls dissipation estimates went down from 0.16-0.18-0.19 to 0.019-0.035-0.053
 - midwater trawls dissipation estimates went up from 0.058-0.07-0.082 to 0.0002-0.105-0.434 (noted huge posterior)
 - also two new dissipation types (tuna vs non-tuna purse seine; beach seine)
"""
import unittest
from ..model_library import GearModelLibrary, MODELS_DIR
from ..base_models import DiscreteChoiceRequired


class TestGearLibrary(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.gml = GearModelLibrary()

    def test_0_num_quantities(self):
        self.assertEqual(len(list(self.gml.quantities('catch'))), 2)
        self.assertEqual(len(list(self.gml.quantities())), 21)

    def test_0_synonym(self):
        self.assertEqual(self.gml.get_quantity('Vessel length in meters'), self.gml.get_quantity('LOA'))

    def test_1_load_models(self):
        self.gml.load_path(MODELS_DIR)
        self.assertEqual(len(list(self.gml.effort_models())), 10)
        self.assertEqual(len(list(self.gml.gear_models())), 13)
        self.assertEqual(len(list(self.gml.dissipation_models())), 25)  # new Gilman prioritization

    def test_2_compose(self):
        k = list(self.gml.valid_models(gear_types={'GFWClass': 'trawlers'}))
        self.assertEqual(len(k), 9)
        with self.assertRaises(DiscreteChoiceRequired):
            [j.mean() for j in k]

    def test_3_mean(self):
        k = list(self.gml.valid_models(gear_types={'GFWClass': 'trawlers'}, gear_family='Watanabe'))
        self.assertEqual(len(k), 3)
        self.assertSetEqual({7.55625, 25.59375, 8.53125}, set(j.mean() for j in k))  # new Gilman prioritization


if __name__ == '__main__':
    unittest.main()



from synonym_dict import SynonymSet, SynonymDict

MEASURES = ('catch', 'scaling', 'operation')  # note: ad hoc measures of specific gear or dissipation models are not managed


class Quantity(SynonymSet):
    def __init__(self, measure, name, unit, *args):
        assert measure in MEASURES
        self._measure = measure
        super(Quantity, self).__init__(name, *args)
        self._unit = str(unit)

    @property
    def unit(self):
        return self._unit

    @property
    def object(self):
        return self

    @property
    def measure(self):
        return self._measure


class QuantityGroup(SynonymDict):
    _entry_group = 'Quantities'
    _syn_type = Quantity

    _ignore_case = True

    _measure = None

    def __init__(self, measure, **kwargs):
        assert measure in MEASURES
        self._measure = measure
        super(QuantityGroup, self).__init__(**kwargs)

    @property
    def group(self):
        return self._measure

    def new_entry(self, *args, **kwargs):
        super(QuantityGroup, self).new_entry(self._measure, *args, **kwargs)



"""
The point of this class is two fold:
 - to collect and organize a set of UnitGearModels
 - to gather fisheries input data and marshall it together to run the models and generate results
 * related to the second goal is to handle stochastic simulations

The libraries themselves are de-serialized simply, but they must be *registered* with the model, including
with any model-specific designations for applicable fisheries or countries
"""

import json
import os
# from collections import defaultdict

from .quantities import QuantityGroup, MEASURES
from .gear_mapping import validate_gear_types
from .stages import CatchEffort, GearIntensity, Dissipation
from .query import GearModel, ConflictingUnits  # , ConflictingParams, NoValidParams  # , ZeroValuedModel


REFERENCE_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'reference'))
MODELS_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'models'))

MODEL_STAGES = ('effort', 'gear', 'dissipation')


class InvalidQuantity(Exception):
    pass


class GearModelLibrary(object):

    def _print(self, st, verbose=None):
        if verbose is None:
            verbose = self._verbose
        if verbose:
            print(st)

    def _load(self, lib_path, overwrite=False, verbose=None):
        # First load quantities
        if lib_path not in self._paths:
            self._paths.append(lib_path)
        self._load_quantities(lib_path, verbose=verbose)
        if len(list(self.quantities())) == 0:
            print('Note: no quantities found in path %s' % lib_path)
        for f in os.listdir(lib_path):
            family, ext = os.path.splitext(f)
            if ext.lower() != '.json':
                continue
            if family.lower() == 'quantities':
                continue
            try:
                self.load_family(lib_path, family, overwrite=overwrite, verbose=verbose)
            except json.JSONDecodeError:
                continue

    def load_family(self, lib_path, family, overwrite=False, verbose=None):
        if family.lower() == 'quantities':
            raise AttributeError('quantities.json is not a data family')
        with open(os.path.join(lib_path, family + '.json')) as fp:
            j = json.load(fp)
        source_doc = '\n'.join('%s: %s' %(k, v) for k, v in j.items() if k.startswith('source_'))
        self._load_effort(family, source_doc, j.pop('effort_models', []), overwrite=overwrite, verbose=verbose)
        self._load_gear(family, source_doc, j.pop('gear_models', []), overwrite=overwrite, verbose=verbose)
        self._load_diss(family, source_doc, j.pop('dissipation_models', []), overwrite=overwrite, verbose=verbose)

    def load_path(self, lib_path, overwrite=True, verbose=True):
        self._load(lib_path, overwrite=overwrite, verbose=verbose)

    def _load_quantities(self, lib_path, verbose=None):
        try:
            with open(os.path.join(lib_path, 'quantities.json')) as fp:
                j = json.load(fp)
        except FileNotFoundError:
            self._print('No quantities file in %s' % lib_path, verbose=verbose)
            return
        qdict = j.pop('quantities', [])
        if len(qdict) == 0:
            return
        for m in MEASURES:
            for qm in qdict.pop(m, []):
                self._print('Adding %s quantity %s' % (m, qm[0]), verbose=verbose)
                self._q[m].new_entry(*qm)

    def _load_effort(self, family, source_doc, entries, overwrite=False, verbose=None):
        for entry in entries:
            self._print('Adding effort model %s' % entry['name'], verbose=verbose)
            entry['catch_unit'] = self.get_quantity(entry.pop('catch_unit'), measure='catch')
            entry['scaling_unit'] = self.get_quantity(entry.pop('scaling_unit'), measure='scaling')
            entry['op_unit'] = self.get_quantity(entry.pop('op_unit'), measure='operation')
            e = CatchEffort(family, source_doc=source_doc, **entry)
            self.add_effort_model(e, overwrite=overwrite)

    def add_effort_model(self, e, overwrite=False):
        assert isinstance(e, CatchEffort)
        self._add_model(e, self._e, overwrite=overwrite)

    def _load_gear(self, family, source_doc, entries, overwrite=False, verbose=None):
        for entry in entries:
            self._print('Adding gear intensity model %s' % entry['name'], verbose=verbose)
            entry['scaling_unit'] = self.get_quantity(entry.pop('scaling_unit'), measure='scaling')
            g = GearIntensity(family, source_doc=source_doc, **entry)
            self.add_gear_model(g, overwrite=overwrite)

    def add_gear_model(self, g, overwrite=False):
        assert isinstance(g, GearIntensity)
        self._add_model(g, self._g, overwrite=overwrite)

    def _load_diss(self, family, source_doc, entries, overwrite=False, verbose=False):
        for entry in entries:
            self._print('Adding dissipation model %s' % entry['name'], verbose=verbose)
            entry['op_unit'] = self.get_quantity(entry.pop('op_unit'), measure='operation')
            d = Dissipation(family, source_doc=source_doc, **entry)
            self.add_dissipation_model(d, overwrite=overwrite)

    def add_dissipation_model(self, d, overwrite=False):
        assert isinstance(d, Dissipation)
        self._add_model(d, self._d, overwrite=overwrite)

    @staticmethod
    def _add_model(m, target, overwrite=False):
        if m.name in target:
            if overwrite:
                print('Overwriting model %s' % m.name)
            else:
                print('Not overwriting model %s' % m.name)
                return
        target[m.name] = m

    def __init__(self, *paths, verbose=False, **kwargs):
        """

        :param gears_path: folder containing
        :param kwargs:

        """
        self._verbose = verbose

        self._q = {m: QuantityGroup(m) for m in MEASURES}
        self._args = kwargs
        self._e = dict()
        self._g = dict()
        self._d = dict()
        self._paths = []
        self._load(REFERENCE_DIR, verbose=verbose)
        for lib_path in paths:
            self._load(lib_path, verbose=verbose)

    def _yield_from_meas(self, meas):
        for k in sorted(self._q[meas].objects, key=lambda x: x.name):
            yield self._q[meas][k]

    def quantities(self, measure=None):
        if measure is not None:
            if measure in self._q:
                for q in self._yield_from_meas(measure):
                    yield q
            else:
                yield ValueError('Unknown measure %s' % measure)
        else:
            for meas in MEASURES:
                for q in self._yield_from_meas(meas):
                    yield q

    def get_quantity(self, name, measure=None):
        q = None
        if measure is None or measure not in MEASURES:
            for v in self._q.values():
                try:
                    q = v[name]
                except KeyError:
                    continue
        else:
            q = self._q[measure][name]
        if q is None:
            raise InvalidQuantity(name)
        return q

    @property
    def effort_names(self):
        for k in sorted(self._e.keys()):
            yield k

    @property
    def gear_names(self):
        for k in sorted(self._g.keys()):
            yield k

    @property
    def dissipation_names(self):
        for k in sorted(self._d.keys()):
            yield k

    @staticmethod
    def _check_model(mod, ec=None, family=None):
        if ec is not None:
            if not bool(mod.gear_ecs.intersection(ec)):
                return False
        if family is not None:
            if not mod.family.lower().startswith(family.lower()):
                return False
        return True

    def effort_models(self, ec=None, family=None, effort=None):
        try:
            v = self._e[effort]
            if self._check_model(v, ec=ec, family=family):
                yield v
        except KeyError:
            for k in self.effort_names:
                v = self._e[k]
                if self._check_model(v, ec=ec, family=family):
                    yield v

    def gear_models(self, ec=None, family=None, gear=None):
        try:
            v = self._g[gear]
            if self._check_model(v, ec=ec, family=family):
                yield v
        except KeyError:
            for k in self.gear_names:
                v = self._g[k]
                if self._check_model(v, ec=ec, family=family):
                    yield v

    def dissipation_models(self, ec=None, family=None, dissipation=None):
        try:
            v = self._d[dissipation]
            if self._check_model(v, ec=ec, family=family):
                yield v
        except KeyError:
            for k in self.dissipation_names:
                v = self._d[k]
                if self._check_model(v, ec=ec, family=family):
                    yield v

    def valid_models(self, gear_types, verbose=None,
                     effort=None, gear=None, dissipation=None,
                     effort_family=None, gear_family=None, dissipation_family=None):
        """
        Generate valid triples of (catch-effort, gear-intensity, dissipation) for the specified gear and parameters.
        {effort, gear, dissipation} = strict matching to name of desired model (still subject to gear type matching)
        {effort_family, gear_family, dissipation_family} = filter when family name starts with supplied string
        :param gear_types: a dict of {family: gear} or {family: [gear1, ...]}
        :param verbose:
        :param effort: literal name to only match
        :param gear: "
        :param dissipation: "
        :param effort_family: Filter by <family>.lower()startswith(<arg>.lower())
        :param gear_family: "
        :param dissipation_family: "
        :return:
        """
        ecs = set(validate_gear_types(gear_types))

        count = 0

        for eff in self.effort_models(ecs, family=effort_family, effort=effort):
            for gm in self.link_effort_model(eff, gear=gear, dissipation=dissipation,
                                             gear_family=gear_family, dissipation_family=dissipation_family,
                                             verbose=verbose, _ecs=ecs):
                self._print('%d: %s-%s-%s passed' % (count, eff, gm.gear, gm.dissipation), verbose=verbose)
                yield gm
                count += 1

    def link_effort_model(self, eff, gear=None, dissipation=None,
                          gear_family=None, dissipation_family=None, verbose=None, _ecs=None):
        """
        This allows externally-specified effort models to use library gear+dissipation models
        :param eff:
        :param gear:
        :param dissipation:
        :param gear_family:
        :param dissipation_family:
        :param verbose:
        :param _ecs: pass a more limiting set of
        :return:
        """
        if _ecs is None:
            _ecs = eff.gear_ecs
        for gea in self.gear_models(_ecs, family=gear_family, gear=gear):
            for dis in self.dissipation_models(_ecs, family=dissipation_family, dissipation=dissipation):
                try:
                    gm = GearModel(eff, gea, dis)
                except ConflictingUnits as e:
                    err = '%s: %s' % (e.__class__.__name__, e)
                    self._print('%s-%s-%s ... failed %s' % (eff, gea, dis, err), verbose=verbose)
                    continue
                yield gm

    @staticmethod
    def models_report(models, e_param=None, g_param=None, d_param=None):
        """
        :param models: an iterable of models, e.g. the output of valid_models
        :param e_param: effort param value used for report generation
        :param g_param: gear ""
        :param d_param: dissipation ""
        :return:
        """
        for mod in models:
            for report in mod.report(e_param=e_param, g_param=g_param, d_param=d_param):
                yield report



"""
map between gear models
"""

import csv
import os
from collections import defaultdict
# from itertools import chain

MASTER_GEAR_FILE = os.path.join(os.path.dirname(__file__), '..', 'master_gear_mapping.csv')


class GearMapper(object):

    def _add_record(self, record):
        this = len(self._classes)
        self._classes.append(record)
        for k, v in record.items():
            v = v.strip()
            if k not in self._sectors:
                self._sectors[k] = list()
            if v not in self._sectors[k]:
                self._sectors[k].append(v)
            self._reverse[k, v].add(this)

    def __init__(self):
        self._sectors = dict()
        self._reverse = defaultdict(set)
        self._classes = []  # I want a dict, but like, where the keys are just automatically increasing whole numbers...
        with open(MASTER_GEAR_FILE) as f:
            _cr = csv.DictReader(f)
            for record in _cr:
                self._add_record(record)

    def _gather_eqs(self, d1):
        first = []
        for k, v in d1.items():
            if isinstance(v, str):
                first.extend(list(self._reverse[k, v]))
            else:
                for t in v:
                    first.extend(list(self._reverse[k, t]))
        return set(first)

    def overlap(self, d1, d2):
        """
        Generates equivalence classes where the two gear types overlap
        :param d1: a valid gear_types dict
        :param d2: a valid gear_types dict
        :return:
        """
        f1 = self._gather_eqs(d1)
        f2 = self._gather_eqs(d2)
        for x in f1.intersection(f2):
            yield x

    def test_equivalence(self, family, gear, mappings):
        """
        Returns a dict containing synonyms in mappings for family: gear
        :param family:
        :param gear:
        :param mappings: dict of family:gear specs
        :return:
        """
        is_true = dict()
        for k, v in mappings.items():
            if v in self.translate(family, gear, k):
                is_true[k] = v
        return is_true

    def validate_gear_type(self, family, gear):
        """

        :param family: must be a column heading in master_gear_mapping
        :param gear: must either be a string found in the column, or an iterable of strings found in the column
        :return:
        """
        sect = self._sectors[family]  # or key error
        if isinstance(gear, str):
            if gear in sect:
                return self._reverse[family, gear]
            raise ValueError('%s: %s|' % (family, gear))
        # assume gear is a list-like
        ecs = set()
        for g in gear:
            if g not in sect:
                raise ValueError('%s: %s|' % (family, g))
            for k in self.c_ids(family, g):
                ecs.add(k)
        return sorted(list(ecs))

    def translate(self, family, gear, query_family):
        """
        Given gear:family (e.g. FAOGearName:Handlines and pole-lines (mechanized)) provide a list of synonyms in
        a target family.
        :param family:
        :param gear:
        :param query_family:
        :return:
        """
        if query_family not in self.families:
            raise KeyError(query_family)
        if gear not in self._sectors[family]:
            raise ValueError(gear)

        s = set(self._classes[k][query_family] for k in self._reverse[family, gear])
        return s

    def sectors(self, family):
        return self._sectors[family]

    def equivs(self, family, gear):
        for c in self._reverse[family, gear]:
            yield self._classes[c]

    def c_ids(self, family, gear):
        for c in self._reverse[family, gear]:
            yield c

    @property
    def families(self):
        for k in self._sectors.keys():
            yield k


gear_mapper = GearMapper()


def validate_gear_types(gear_types):
    if not isinstance(gear_types, dict):
        raise TypeError('Gear Type specification must be a dict')
    c_ids = []
    for k, v in gear_types.items():
        c_ids.extend(gear_mapper.validate_gear_type(k, v))
    return sorted(list(set(c_ids)))


def test_overlap(g1, g2):
    try:
        next(gear_mapper.overlap(g1, g2))
    except StopIteration:
        return False
    return True



"""
Stages have a lot of features in common. They all [each] have models, for instance.

Note that a "unit of effort" is composed of [scaling unit] * [operation unit]


Common to all models:
 * name - basically a unique identifier-- should I make an external_ref as well??? ans: not internally
 * gear_types - a dict of Family: Entry or Family: [entry1, entry2..] that passes validate_gear_types()
 * model: a valid model
 - documentation - free text
 - ref_unit - a reference unit for the input flow. This is catch_unit for effort models, scaling_unit for gear models,
   and op_unit for dissipation models.
 - param_unit: quantity of the model input parameter. Must be supplied if model order > 0. can be any quantity.
 - param_max: input parameter value above which the model is invalid
 - param_min: input parameter value below which the model is invalid

Effort Models also specify downstream outputs:
 + catch_unit: a quantity measuring catch
 + scaling_unit: a quantity measuring the scale of the operation- may or may not be the same as param_unit
 + op_unit: a quantity measuring operation
 + op_equiv: a dict of { quantity: amount of that quantity that is equivalent to 1 op_unit}
 : model maps catch in catch_unit to effort in scaling_unit * op_unit
 : op_equiv example: op_unit is 1 fishing day, op_equiv could be 6 fishing hours or 0.01 years
 : op_factor(unit) returns the op_equiv directly, or 1.0 for unit == op_unit. this is multiplied by the result
   This is because we are converting op_unit to op_equiv

Gear Models add:
 + scaling_unit: a quantity measuring the scale of the operation- may or may not be the same as param_unit
 + attributes: dict of { ad hoc gear measure: amount per kg_gear }
 : model maps scaling_unit to kg_gear
 : if param_unit is omitted but required, it is assumed to be scaling_unit

Dissipation Models add:
 + op_unit: a quantity measuring operation
 + dissipation_type: ad hoc dissipation measure
 + op_equiv: a dict of { quantity: amount of that quantity that is equivalent to 1 op_unit
 : model maps op_unit to fraction dissipation
 : if param_unit is omitted but required, it is assumed to be op_unit
 : op_factor(unit) returns the inverse of op_equiv, or 1.0 for unit == op_unit. this is multiplied by the result.
   This is because we are converting op_equiv to op_unit


"""
from .base_models import PolynomialModel, DiscreteModel, DiscreteChoiceRequired, BaseModel
from .gear_mapping import validate_gear_types

from random import choice

class ModelAlreadyDefined(Exception):
    pass


class ModelStage(object):
    stage = None
    _model = None
    ref_quantity = None

    def __init__(self, family, name, gear_types, model, _equiv, param_unit=None, param_min=None, param_max=None,
                 model_scale=None,
                 documentation=None, source_doc=None):
        """

        :param family:
        :param name:
        :param gear_types:
        :param model:
        :param _equiv:
        :param param_unit:
        :param param_min:
        :param param_max:
        :param model_scale:
        :param documentation:
        :param source_doc:
        """

        self.family = family
        self.name = name
        self.gear_ecs = set(validate_gear_types(gear_types))
        self.gear_types = gear_types

        if param_unit is None:
            if self.ref_quantity is not None:
                param_unit = getattr(self, self.ref_quantity)

        self.param_unit = param_unit
        self.param_min = param_min
        self.param_max = param_max
        self.model_scale = model_scale

        if _equiv is None:
            _equiv = dict()
        _equiv[self._equiv_unit.name] = 1.0

        self._equiv = _equiv

        self._doc = '\n'.join([(source_doc or "(No source documentation)"), documentation])

        self._model = self._parse_arg(model)
        if self._model.order > 0 and self.param_unit is None:
            print('%s: Warning: unspecified unit for required input parameter' % self.name)

    def __repr__(self):
        return '%s(%s: %d%s)[%s]'% (self.__class__.__name__, self.name, self.order, self.model_type[0], self.family)

    @property
    def _equiv_unit(self):
        return self.ref_unit

    def _get_cf(self, key):
        try:
            return next(v for k, v in self._equiv.items() if key == k)
        except StopIteration:
            raise KeyError(key)

    @property
    def ref_unit(self):
        return getattr(self, self.ref_quantity)

    @property
    def doc(self):
        return self._doc

    @property
    def model_type(self):
        if isinstance(self._model, DiscreteModel):
            return 'discrete'
        elif isinstance(self._model, PolynomialModel):
            return 'continuous'
        return 'null'

    @property
    def order(self):
        return self._model.order

    def mean(self, param=None):
        """
        Returns the determining parameter value and the mean
        :param param:
        :return: param, value
        """
        values = list(self.values(param))
        if len(values) > 1:
            print(self.name)
            raise DiscreteChoiceRequired('Ambiguous param specification: %s' % values)
        elif len(values) == 0:
            return None, self._model.mean(None)
        else:
            return values[0], self._model.mean(values[0])

    def sample(self, param=None):
        values = list(self.values(param))
        value = choice(values)
        # if self.model_type == 'continuous':
        #     if self.order > 0 and value == 0.0:
        #         # this warning is to prevent silent param-lookup fails on continuous models
        #         print('Warning: sampling linear model with 0.0 param %s' % self.name)
        return value, self._model.sample(value)

    @property
    def keys(self):
        if self.model_type == 'discrete':
            for v in self._model.valid_params:
                yield v
        elif self.model_type == 'continuous':
            yield self.param_unit

    def values(self, param):
        """
        Generates all valid parameter values consistent with the input argument.

        For Continuous model types: Param can be:
         - None -> for 0-order models, yield 0.0
         - a number -> interpreted as a number in param_unit
         - a set, list, or tuple of numbers -> each generated
         - a dict of {unit: number} pairs -> passed into _equiv and converted
         - a dict of {unit: tuple} pairs -> each item passed into _equiv and converted

         The general case is a tuple of numbers expressed in terms of param_unit

        For Discrete model types: Param can be:
         - None -> generate keys
         - a string -> generate if string is a valid key
         - a list, set, or tuple of strings -> generate intersection with keys
         - a dict of {key: bool} -> generate intersection of True keys with keys

         The general case is a tuple of members of keys

        This is brilliant but deserves maybe a modicum of testing.

        note: obviously this is terrible.

        :param param:
        :return:
        """
        if self.model_type == 'continuous':
            if isinstance(param, set) or isinstance(param, list) or isinstance(param, tuple):
                _gen = (float(k) for k in param if self.is_valid(k))
            elif isinstance(param, dict):
                if self._equiv_unit == self.param_unit:  # conversion is valid
                    _gen = []
                    for k, v in param.items():
                        try:
                            factor = self._get_cf(k)
                        except KeyError:
                            continue
                        _gen.extend([t / factor for t in self.values(v)])
                else:
                    try:
                        _gen = self.values(param[self.param_unit])
                    except KeyError:
                        _gen = ()
            else:
                _gen = (k for k in (param, ) if self.is_valid(k))
        elif self.model_type == 'discrete':
            if param is None:
                _gen = self.keys
            else:
                if isinstance(param, dict):
                    param = [k for k, v in param.items() if bool(v)]
                if isinstance(param, set) or isinstance(param, list) or isinstance(param, tuple):
                    _gen = (k for k in param if self.is_valid(k))
                else:
                    _gen = (k for k in (param, ) if self.is_valid(k))
        else:
            _gen = ()

        for p in _gen:
            yield p

    def is_valid(self, param_value=None):
        """
        For qualitative models, 'param_value' must be in self.keys
        for quantitative models, if order > 0, value must be in bounds
        :param param_value:
        :return:
        """
        if self.model_type == 'null':
            return False
        if self.model_type == 'continuous':
            if self._model.order == 0:
                return True
            else:
                try:
                    value = float(param_value)
                except (TypeError, ValueError):
                    return False
                if self.param_min is not None:
                    if value < self.param_min:
                        return False
                if self.param_max is not None:
                    if value > self.param_max:
                        return False
                    return True
                return True
        if self.model_type == 'discrete':
            return param_value in self._model.valid_params
        return False

    def _parse_arg(self, arg):
        """
        Three possibilities:
         - a single number
         - an iterable whose first entry is a string (uncertainty spec)
         - an iterable whose first entry is a non-string iterable
        :param arg:
        :return: either arg
        """
        if isinstance(arg, BaseModel):
            return arg
        if self.model_type != 'null':
            raise ModelAlreadyDefined()
        if arg is None:
            arg = 1.0
        if isinstance(arg, dict):
            dm = DiscreteModel(**arg)
            return dm
        try:
            first = arg[0]
        except TypeError:
            # not indexable- just a number
            return PolynomialModel(arg, scale=self.model_scale)
        if isinstance(first, str):
            # uncertainty spec
            return PolynomialModel(arg, scale=self.model_scale)
        # higher-order model
        return PolynomialModel(*arg, scale=self.model_scale)


class CatchEffort(ModelStage):
    stage = 'effort'

    @property
    def ref_unit(self):
        return self.catch_unit

    def __init__(self, family, name, gear_types, catch_unit, scaling_unit, op_unit, effort_model, op_equiv=None, **kwargs):
        self.catch_unit = catch_unit
        self.scaling_unit = scaling_unit
        self.op_unit = op_unit

        super(CatchEffort, self).__init__(family, name, gear_types, effort_model, op_equiv, **kwargs)


    @property
    def _equiv_unit(self):
        return self.op_unit

    @property
    def op_units(self):
        """
        generates operation units
        :return:
        """
        for k in self._equiv.keys():
            yield k

    def op_factor(self, op_unit):
        return self._get_cf(op_unit)


class GearIntensity(ModelStage):
    stage = 'gear'
    ref_quantity = 'scaling_unit'

    def __init__(self, family, name, gear_types, scaling_unit, intensity_model, scaling_equiv=None,
                 attributes=None, **kwargs):
        self.scaling_unit = scaling_unit
        super(GearIntensity, self).__init__(family, name, gear_types, intensity_model, scaling_equiv, **kwargs)

        if attributes is None:
            attributes = dict()
        self.attributes = attributes

    @property
    def scaling_units(self):
        """
        generates operation units
        :return:
        """
        for k in self._equiv.keys():
            yield k

    def scaling_factor(self, scaling_unit):
        return self._get_cf(scaling_unit)


class Dissipation(ModelStage):
    stage = 'dissipation'
    ref_quantity = 'op_unit'

    def __init__(self, family, name, gear_types, op_unit, dissipation_type, dissipation_model, op_equiv=None, **kwargs):
        self.op_unit = op_unit
        self.dissipation_type = dissipation_type
        super(Dissipation, self).__init__(family, name, gear_types, dissipation_model, op_equiv, **kwargs)

    @property
    def op_units(self):
        """
        generates operation units
        :return:
        """
        for k in self._equiv.keys():
            yield k

    def op_factor(self, op_unit):
        return self._get_cf(op_unit)



"""
A model is something that converts an input to an output. Our base model is a generic n-order model of one variable,
for which each parameter can be either static or uncertain.
"""

from stats_arrays import (UncertaintyBase, MCRandomNumberGenerator, NormalUncertainty, LognormalUncertainty, \
    UniformUncertainty, TriangularUncertainty, NoUncertainty)

from math import log, exp


class EmptyModel(Exception):
    """
    Models should have at least one parameter
    """
    pass


class BaseModel(object):
    """
    Abstract model class
    """
    @property
    def order(self):
        raise NotImplementedError

    def mean(self, param):
        raise NotImplementedError

    def sample(self, param):
        raise NotImplementedError

    @property
    def valid_params(self):
        raise NotImplementedError


class InoperableModel(Exception):
    """
    Discrete models must be 0-order because they cannot take parameters

    This should not apply to multimodal continuous models [not yet implemented] where the param values are numeric
    and the model selection param is also used as the model input. In these cases, the keys would be ordered, and the
    model selected would be the least-greater-valued key.  The highest-valued key thus signifies an upper bound for
    the param.
    """
    pass


class PolynomialModel(BaseModel):
    _log = False
    _log10 = False

    def _make_stats_array_input(self, arg):
        """
        Strictly speaking, the order params for triangular distributions are specified is irrelevant, since they can
        be sorted.
        This is also true for uniform distributions.
        :param arg:
        :return:
        """
        try:
            typ = arg[0].lower()[0]  # ignore typos (and i18n) with this simple trick!
            mean = float(arg[1])
            if typ == 'n':
                d = {'loc': float(arg[1]), 'scale': float(arg[2]), 'uncertainty_type': NormalUncertainty.id}
                if self.bounded:
                    # bound normal distributions on the same side of 0 for stability purposes
                    if d['loc'] > 0:
                        d['minimum'] = d['loc'] / 1e3
                    elif d['loc'] < 0:
                        d['maximum'] = d['loc'] / 1e3
            elif typ == 'l':  # take log of the data
                d = {'loc': log(float(arg[1])), 'scale': float(arg[2]), 'uncertainty_type': LognormalUncertainty.id}
            elif typ == 'u':
                mn, mx = tuple(sorted(float(a) for a in arg[1:]))
                d = {'maximum': mx, 'minimum': mn, 'uncertainty_type': UniformUncertainty.id}
                mean = (mn + mx) / 2
            elif typ == 't':
                mn, lc, mx = tuple(sorted(float(a) for a in arg[1:]))
                d = {'loc': lc, 'maximum': mx, 'minimum': mn, 'uncertainty_type': TriangularUncertainty.id}
            elif typ == 's':
                d = {'loc': float(arg[1]), 'uncertainty_type': NoUncertainty.id}
            else:
                raise ValueError('Unknown uncertainty type %s' % arg[0])
        except (TypeError, IndexError):
            d = {'loc': float(arg), 'uncertainty_type': NoUncertainty.id}
            mean = float(arg)

        self._params.append(d)
        self._means.append(mean)

    @staticmethod
    def _p_str(k):
        return {
            NoUncertainty.id: '',
            NormalUncertainty.id: ' N',
            LognormalUncertainty.id: ' L',
            TriangularUncertainty.id: ' T',
            UniformUncertainty.id: ' U'
        }[k['uncertainty_type']]

    @staticmethod
    def _re_parse_args(args):
        args = list(args)
        orders = []
        while len(args) > 0:
            coef = [args.pop(0)]
            while len(args) > 0 and not isinstance(args[0], str):
                coef.append(args.pop(0))
            orders.append(tuple(coef))
        return orders

    def __init__(self, *args, bounded=True, scale=None):
        """
        Generate a polynomial model, with or without uncertainty.  The number of positional args determines the
        order of the model.

        I'm sure someone has done this better.

        To implement an uncertain model on *any* argument, the argument should be a tuple with the following format:
         ('normal', mean, stdev)  # note: will be bounded at 0
         ('lognormal', exp(mu), sigma)  # note: enter the true value of the mean, not mean of the underlying normal dist.
         ('uniform', high, low)
         ('triangular', mode, high, low)
         ('static', mean) # no uncertainty

        :param args:
        :param scale: None or default: polynomial models are linear in the parameter
         'log': Model applies to log-transformed data; the computation returns exp(result)
         'log10': Model applies to log10-transformed data; the computation returns 10**(result).
        :param bounded: [True] bound normal distributions on the same side of 0 as loc, for stability purposes.
        """
        if len(args) == 0:
            raise EmptyModel()
        if isinstance(args[0], str):
            # 0-order uncertain model specified with positional params
            args = self._re_parse_args(args)
        self.scale = scale
        self._params = []
        self._means = []
        self.bounded = bounded
        for arg in args:
            self._make_stats_array_input(arg)

        self._mcg = MCRandomNumberGenerator(UncertaintyBase.from_dicts(*self._params))

    @property
    def valid_params(self):
        return []

    @property
    def scale(self):
        if self._log:
            return 'log'
        elif self._log10:
            return 'log10'
        else:
            return 'linear'

    @scale.setter
    def scale(self, value):
        if value is None:
            return
        if value == 'log':
            self._log = True
            self._log10 = False
        elif value == 'log10':
            self._log10 = True
            self._log = False
        elif value == 'linear':
            self._log = self._log10 = False
        if self._log and self._log10:
            self._log = self._log10 = False
            raise ValueError('Cannot specify both log and log10')  # this should never happen

    @property
    def order(self):
        return len(self._params) - 1

    def _compute(self, x, arr):
        x = x or 0.0
        y = arr[0]
        for i in range(self.order):
            y += arr[i+1] * x**(i+1)
        if self._log:
            return exp(y)
        elif self._log10:
            return 10 ** y
        return y

    def mean(self, value=0.0):
        return self._compute(value, self._means)

    def sample(self, value=0.0):
        res = [float(k) for k in next(self._mcg)]
        return self._compute(value, res)

    @property
    def y_int(self):
        return self.mean(0.0)

    @property
    def _coef(self):
        return ';'.join(['%d:%.3g%s' % (i, self._means[i], self._p_str(k)) for i, k in enumerate(self._params)])

    @property
    def _exp(self):
        if self._log:
            return 'exp '
        elif self._log10:
            return '10 ** '
        else:
            return ''

    def __str__(self):
        return 'y ~ %s%s (x)' % (self._exp, self._coef)

    def __repr__(self):
        return '%s(%s%s)' % (self.__class__.__name__, self._exp, self._coef)


class DiscreteChoiceRequired(Exception):
    pass


class DiscreteModel(BaseModel):
    """
    A mapping of discrete params to 0-order models, with or without uncertainty. In a discrete model, a qualitative
    param (dict lookup) is used to select from a range of alternative models. The models must have 0-order because
    at the moment there is no mechanism to pass a second param to be used as the model input.
    """
    def __init__(self, **params):
        self._models = dict()
        for param, mdl in params.items():
            try:
                model = PolynomialModel(mdl)
            except AttributeError:
                raise InoperableModel('params: %s' % str(mdl))
            if model.order != 0:
                raise InoperableModel(param)
            self._models[param] = model

    @property
    def order(self):
        return 0

    @property
    def valid_params(self):
        return tuple(sorted(self._models.keys()))

    def mean(self, param):
        try:
            return self._models[param].mean()
        except KeyError:
            raise DiscreteChoiceRequired(self.valid_params)

    def sample(self, param):
        try:
            return self._models[param].sample()
        except KeyError:
            raise DiscreteChoiceRequired(self.valid_params)

    @property
    def _means(self):
        return '; '.join(['%s: %.3g' % (k, self.mean(k)) for k in self.valid_params])

    def __str__(self):
        return 'y : %s (x)' % self._means

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self._means)



import unittest

from ..gear_mapping import validate_gear_types


class GearModelTest(unittest.TestCase):
    def test_valid_gear_type(self):
        self.assertTrue(validate_gear_types({'GFWClass': 'squid_jigger'}))

    def test_valid_gear_types(self):
        self.assertTrue(validate_gear_types({'VonBrandt3Code': ('7', '9.2', '5.6.3')}))

    def test_invalid_gear_type(self):
        with self.assertRaises(KeyError):
            validate_gear_types({'VonBrundtland': '42'})
        with self.assertRaises(ValueError):
            validate_gear_types({'VonBrandtDescription': ('7', '9.2', '5.6.3')})

    def test_fads(self):
        self.assertTrue(validate_gear_types({'FADs': "Purse Seine Anchored FADs"}))


if __name__ == '__main__':
    unittest.main()



import unittest

from ..base_models import PolynomialModel, DiscreteModel, InoperableModel, EmptyModel


class BaseModelsTest(unittest.TestCase):
    """
    Note that we are not implementing any tests as to the correctness of the distributions
    """
    def test_static_0_order(self):
        sm = PolynomialModel(43.2)
        self.assertEqual(sm.mean(), 43.2)
        self.assertEqual(sm.sample(), 43.2)

    def test_uncertain_0_order_strict(self):
        sm = PolynomialModel(('normal', 17.5, 2.3))
        self.assertEqual(sm.order, 0)
        self.assertEqual(sm.mean(), 17.5)
        self.assertNotEqual(sm.sample(), 17.5)

    def test_uncertain_1_order_sloppy(self):
        a0, a0s = 33.2, 1.1
        a1, a1s = 0.54, 0.02
        a1mean = (a1 + a1s) / 2

        sm = PolynomialModel('normal', a0, a0s, 'uniform', a1, a1s)
        self.assertEqual(sm.order, 1)
        self.assertEqual(sm.mean(), a0)
        self.assertEqual(sm.mean(35), a0 + 35*a1mean)
        self.assertNotEqual(sm.sample(333), sm.mean(333))

    def test_static_higher_order(self):
        ps = [1, 2.3, 0.456, 0.789]
        sm = PolynomialModel(*ps)
        x = 4.2
        y = 0.0
        for i in range(len(ps)):
            y += ps[i]*x**i
        self.assertEqual(sm.mean(x), y)

    def test_mixed_certain_uncertain(self):
        sm = PolynomialModel(-3.4, ('normal', 7.1, 2.1), .25)
        self.assertEqual(sm.order, 2)
        self.assertEqual(sm.mean(1), -3.4 + 7.1 + .25)
        self.assertEqual(sm.mean(2), -3.4 + 14.2 + 1)
        self.assertNotEqual(sm.sample(1), sm.mean(1))

    def test_init_list(self):
        sm = PolynomialModel(["triangular", 0.066, 0.073, 0.059])
        self.assertEqual(sm.order, 0)

    def test_lognormal_validity(self):
        """
        just confirms that we are specifying the lognormal dist and not the base normal dist
        """
        a0, a1 = 350, 0.000001
        sm = PolynomialModel('lognormal', a0, a1)
        self.assertEqual(sm.mean(), a0)
        self.assertAlmostEqual(sm.mean(), sm.sample(), places=1)

    def test_empty(self):
        with self.assertRaises(EmptyModel):
            PolynomialModel()

    def test_inop(self):
        with self.assertRaises(InoperableModel):
            DiscreteModel(some=(45, 37))

    def test_discrete(self):
        dm = DiscreteModel(some=35, others=45)
        self.assertTupleEqual(dm.valid_params, ('others', 'some'))
        self.assertEqual(dm.mean('others'), 45)
        self.assertEqual(dm.sample('some'), 35)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

requires = [
    'stats_arrays>=0.6.4',
    'synonym_dict>=0.1.1',
    'numpy<2.0'
]

'''
VERSION HISTORY
0.1.1 - 16 June 2024 - improve requirements specification

0.1.0 - 18 November 2020 - Initial setup 
'''

VERSION = '0.1.1'

setup(
    name="unit_gears",
    version=VERSION,
    author="Brandon Kuczenski",
    author_email="bkuczenski@ucsb.edu",
    license=open('LICENSE').read(),
    install_requires=requires,
    url="https://github.com/bkuczenski/unit_gears",
    summary="A library of uncertain models for the industrial operation of fishing gears",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages()
)
